@@ -1,18 +1,19 @@
 # coding: utf-8
-from __future__ import absolute_import, unicode_literals
+from __future__ import absolute_import, unicode_literals, print_function
 
+from FL import fl
 import gc
 import os
 
 resources_path = os.path.join(os.path.dirname(__file__), 'resources')
 if resources_path not in os.environ['PATH']:
-	os.environ['PATH'] += '%s;' % resources_path
+  os.environ['PATH'] += '%s;' % resources_path
 
 from . import core
 
 show_default_optimize_code_points = core.show_default_optimize_code_points
 
-__version__ = '0.8.0'
+__version__ = '0.8.1'
 __doc__ = """
 VFB2UFO3
 DESCRIPTION
@@ -29,6 +30,10 @@ not identifiable as first/second from their name. All glyph hints/links are
 ignored by default; a batch command can be created for use with `psautohint`,
 which supports hinting outlines with decimal coordinates.
 
+VFB2UFO3 also quickly creates UFO masters from multiple master `.vfb` fonts.
+This is the default export option when the `instance_values` option is not
+supplied.
+
 INSTALLATION
 PyPi
 `pip install vfb2ufo3`
@@ -45,33 +50,33 @@ below.
 
 ```
 [user folder]
-    └── Documents
-        └── Fontlab
-            └── Studio 5
-                └── Macros
-                    └── System
-                        └── Modules
-                            └── vfb2ufo3
-                                ├── resources
-                                │   ├── libgcc_s_dw2-1.dll
-                                │   ├── libgomp-1.dll
-                                │   ├── libstdc++-6.dll
-                                │   ├── libwinpthread-1.dll
-                                │   └── zlib1.dll
-                                ├── __init__.py
-                                ├── core.pyd
-                                ├── designspace.pyd
-                                ├── fdk.pyd
-                                ├── fea.pyd
-                                ├── fontinfo.pyd
-                                ├── glif.pyd
-                                ├── groups.pyd
-                                ├── kern.pyd
-                                ├── mark.pyd
-                                ├── plist.pyd
-                                ├── tools.pyd
-                                ├── user.py
-                                └── vfb.pyd
+  └─ Documents
+     └─ Fontlab
+        └─ Studio 5
+           └─ Macros
+              └─ System
+                 └─ Modules
+                    └─ vfb2ufo3
+                       ├─ resources
+                       │  ├─ libgcc_s_dw2-1.dll
+                       │  ├─ libgomp-1.dll
+                       │  ├─ libstdc++-6.dll
+                       │  ├─ libwinpthread-1.dll
+                       │  └─ zlib1.dll
+                       ├─ __init__.py
+                       ├─ core.pyd
+                       ├─ designspace.pyd
+                       ├─ fdk.pyd
+                       ├─ fea.pyd
+                       ├─ fontinfo.pyd
+                       ├─ glif.pyd
+                       ├─ groups.pyd
+                       ├─ kern.pyd
+                       ├─ mark.pyd
+                       ├─ plist.pyd
+                       ├─ tools.pyd
+                       ├─ user.py
+                       └─ vfb.pyd
 ```
 
 REQUIREMENTS
@@ -101,8 +106,8 @@ OPTIONAL
   https://www.msys2.org
   http://mingw.org
 
-  SHA512
-  https://github.com/pr0f3ss/SHA512
+  {fmt}
+  https://github.com/fmtlib/fmt
 
 FUNCTIONALITY
 UFO output is produced without changes to the source font. If a specific
@@ -115,11 +120,6 @@ and/or glyph names and suffixes which can be omitted from the final UFO
 instance (`glyphs_omit_names`, `glyphs_omit_suffixes`). See GLYPHS OPTIONS
 below.
 
-Generated instances/layers can be saved and/or left open after generation via
-the `vfb_save` and `vfb_close` options. If `vfb_save` is set to `True`, the
-resulting `.vfb` instance will be updated during UFO creation. This includes
-glyph outline changes (overlap removal and decomposition).
-
 All path options must be absolute paths; folder and file paths which are not
 absolute will be ignored. The default output path is the user's Desktop.
 
@@ -147,35 +147,35 @@ user_profile_folder = os.environ['USERPROFILE']
 output_path = os.path.join(user_profile_folder, 'Documents', 'test_font')
 
 instances = [
-	0,
-	200,
-	400,
-	650,
-	1000,
-	]
+  0,
+  200,
+  400,
+  650,
+  1000,
+  ]
 names = [
-	'Thin',
-	'Light',
-	'Regular',
-	'SemiBold',
-	'Bold',
-	]
+  'Thin',
+  'Light',
+  'Regular',
+  'SemiBold',
+  'Bold',
+  ]
 attributes = [
-	{'openTypeOS2WeightClass': 200},
-	{'openTypeOS2WeightClass': 300},
-	{'openTypeOS2WeightClass': 400},
-	{'openTypeOS2WeightClass': 600},
-	{'openTypeOS2WeightClass': 700},
-	]
+  {'openTypeOS2WeightClass': 200},
+  {'openTypeOS2WeightClass': 300},
+  {'openTypeOS2WeightClass': 400},
+  {'openTypeOS2WeightClass': 600},
+  {'openTypeOS2WeightClass': 700},
+  ]
 
 write_ufo(
-	output_path=output_path,
-	instance_values=instances,
-	instance_names=names,
-	instance_attributes=attributes,
-	glyphs_decompose=True,
-	glyphs_remove_overlaps=True,
-	)
+  output_path=output_path,
+  instance_values=instances,
+  instance_names=names,
+  instance_attributes=attributes,
+  glyphs_decompose=True,
+  glyphs_remove_overlaps=True,
+  )
 
 -------------------------------------------------------------------------------
 ```
@@ -187,9 +187,7 @@ say 2048, it should be set via the `scale_to_upm` option. The value from the
 turned off by setting `scale_auto` to `False`; this does not reduce conversion
 times.
 
-All scaling operations are performed independently from FontLab; if the `.vfb`
-instance(s)/master copy are being saved, the `.vfb` font and glyph values will
-be the original un-scaled values.
+All scaling operations are performed independently from FontLab.
 
 INSTANCE OPTIONS
 When creating instances from a multiple master source font, lists of values,
@@ -236,8 +234,7 @@ instance_names = ['Light Display']
 ```
 
 GLYPH OPTIONS
-Glyph scaling is independent from the `.vfb` instance itself; if the `.vfb`
-instance is being saved, the glyphs in the `.vfb` will remain un-scaled.
+Glyph scaling is independent from the `.vfb` instance itself.
 
 Glyph decomposition and overlap removal is optional and occurs after instances
 are generated. This option is intended for final output when a binary font will
@@ -250,21 +247,33 @@ contours from each component's base contours.
 When decomposing only, the optimization outlined above will be used for all
 glyphs containing components.
 
-The generated `.vfb` instance(s) will leave components in component-form.
-
 To disable the optimizations outlined above, set the `glyphs_optimize` option
 to `False`.
 
+To specify explicitly which glyphs to decompose and/or remove overlaps, the
+`glyphs_decompose_names` and `glyphs_remove_overlap_names` options should be
+supplied. The supplied names must match exactly with the glyph names in the
+target font. The `glyphs_optimize` option should be disabled if using either of
+these options.
+
 Omit glyphs from instance
   A list of glyph suffixes and/or glyph names can be supplied that should be
   omitted from the instance UFO via the `glyphs_omit_suffixes` and
   `glyphs_omit_names` options, respectively.
 
 Optimize glyph name and code point lists
+  For output intended for makeOTF where construction glyphs are being omitted
+  via `glyphs_omit_list` or `glyphs_omit_suffixes_list`, enabling
+  `glyphs_optimize_makeotf` will decompose and remove overlaps for the omitted
+  glyphs where needed. Overlaps will be removed for all other base components
+  but remain in component-form. This will decrease UFO creation time and allow
+  makeOTF to more efficiently subroutinize the outlines and reduce .otf
+  compilation time.
+
   If removing overlaps, the default list of code points for glyphs to be
   constructed in the above manner is composed of glyphs that normally have no
-  overlapping components. The list is located in the `ufo.pxi` source file with
-  the character representations for these code points shown below.
+  overlapping components. The list is located in the `core.pxi` source file
+  with the character representations for these code points shown below.
 
   The code points from the default code point list can be shown by running the
   `vfb2ufo3.show_default_optimize_code_points()` function. The code points will
@@ -366,19 +375,17 @@ Greek Mono- and Polytonic
 
 Features options
 Font groups can be added to the `features.fea` file on export by setting
-`features_import_groups` to `True`. The font's features are neither formatted
-nor checked for correctness. Users are responsible for moving referenced
-feature files from `include()` statements to the chosen output directory. Also
-see KERN FEATURE OPTIONS and MARK FEATURE OPTIONS below for `kern` and `mark`
-feature options.
+`features_import_groups` to `True` (if `afdko_parts` is enabled, this will also
+be enabled). The font's features are neither formatted nor checked for
+correctness. Users are responsible for moving referenced feature files from
+`include()` statements to the chosen output directory. Also see KERN FEATURE
+OPTIONS and MARK FEATURE OPTIONS below for `kern` and `mark` feature options.
 
 Hint options
 If enabled, the `glyph_hints` option converts hints on export (glyph links are
 converted to hints). By default, hints will be created following the UFO3
 public.postscript.hints specification. The `glyphs_hints_afdko_v1` and
-`glyphs_hints_afdko_v2` options enables hints compatible with MakeOTF. SHA512
-provides the hashing algorithm for computing the glyph hint ids according to
-the UFO3 PostScript hint specification.
+`glyphs_hints_afdko_v2` options enables hints compatible with MakeOTF.
 
 `glyphs_hints_vertical_only` ignored horizontal hints/links
 
@@ -458,27 +465,27 @@ _public.kern2.A: A' AE Agrave Aacute Acircumflex Atilde Adieresis Aring...
 Final output (UFO groups.plist):
 <key>public.kern1.A</key>
 <array>
-	<string>A</string>
-	<string>Agrave</string>
-	<string>Aacute</string>
-	<string>Acircumflex</string>
-	<string>Atilde</string>
-	<string>Adieresis</string>
-	<string>Aring</string>
-	...
+  <string>A</string>
+  <string>Agrave</string>
+  <string>Aacute</string>
+  <string>Acircumflex</string>
+  <string>Atilde</string>
+  <string>Adieresis</string>
+  <string>Aring</string>
+  ...
 </array>
 
 <key>public.kern2.A</key>
 <array>
-	<string>A</string>
-	<string>AE</string>
-	<string>Agrave</string>
-	<string>Aacute</string>
-	<string>Acircumflex</string>
-	<string>Atilde</string>
-	<string>Adieresis</string>
-	<string>Aring</string>
-	...
+  <string>A</string>
+  <string>AE</string>
+  <string>Agrave</string>
+  <string>Aacute</string>
+  <string>Acircumflex</string>
+  <string>Atilde</string>
+  <string>Adieresis</string>
+  <string>Aring</string>
+  ...
 </array>
 ```
 
@@ -507,6 +514,8 @@ There are several explicit keyword options to enable specific MakeOTF switches.
 For those not available via a keyword option, they should be defined as a list
 of strings and passed to the `afdko_makeotf_args` option.
 
+Note: `afdko` for Python 2.7 is no longer developed.
+
 psautohint options
 `psautohint` can be utilized for generating glyph hints after UFO generation.
 Commands to run `psautohint` for each generated instance separately or all
@@ -515,6 +524,8 @@ instances as a batch using the options `psautohint_cmd` and
 decimal (float) hint coordinates) and `-w` (write hints directly to the .glif
 lib for each glyph).
 
+Note: `psautohint` for Python 2.7 is no longer developed.
+
 UFOZ options
 UFO instances can be written as a `.ufoz` archive. If you are planning on any
 file transfer operations after creation, transferring a single `.ufoz` file is
@@ -550,10 +561,10 @@ Test (~3200 glyphs @ 10,000 UPM -> 1,000 UPM), <10 sec
 flc_path = <path to .flc file>
 
 vfb2ufo3.write_ufo(
-	glyphs_decompose=True,
-	glyphs_remove_overlaps=True,
-	groups_flc_path=flc_path,
-	)
+  glyphs_decompose=True,
+  glyphs_remove_overlaps=True,
+  groups_flc_path=flc_path,
+  )
 ```
 
 Test (~2900 glyphs @ 10,000 UPM -> 1,000 UPM), ≈9 sec
@@ -561,19 +572,19 @@ Test (~2900 glyphs @ 10,000 UPM -> 1,000 UPM), ≈9 sec
 ```
 flc_path = <path to .flc file>
 glyphs_omit_list = [
-	<glyph names to be omitted go here>
-	]
+  <glyph names to be omitted go here>
+  ]
 glyphs_omit_suffixes_list = [
-	<glyph name suffixes to be omitted go here>
-	]
+  <glyph name suffixes to be omitted go here>
+  ]
 
 vfb2ufo3.write_ufo(
-	glyphs_decompose=True,
-	glyphs_remove_overlaps=True,
-	glyphs_omit_list=glyphs_omit_list,
-	glyphs_omit_suffixes_list=glyphs_omit_suffixes_list,
-	groups_flc_path=flc_path,
-	)
+  glyphs_decompose=True,
+  glyphs_remove_overlaps=True,
+  glyphs_omit_list=glyphs_omit_list,
+  glyphs_omit_suffixes_list=glyphs_omit_suffixes_list,
+  groups_flc_path=flc_path,
+  )
 ```
 
 Test (~2900 glyphs @ 10,000 UPM -> 1,000 UPM), ≈7 sec
@@ -581,23 +592,23 @@ Test (~2900 glyphs @ 10,000 UPM -> 1,000 UPM), ≈7 sec
 ```
 flc_path = <path to .flc file>
 glyphs_optimize_names = [
-	<glyph names with no overlapping components go here>
-	]
+  <glyph names with no overlapping components go here>
+  ]
 glyphs_omit_list = [
-	<glyph names to be omitted go here>
-	]
+  <glyph names to be omitted go here>
+  ]
 glyphs_omit_suffixes_list = [
-	<glyph name suffixes to be omitted go here>
-	]
+  <glyph name suffixes to be omitted go here>
+  ]
 
 vfb2ufo3.write_ufo(
-	glyphs_decompose=True,
-	glyphs_remove_overlaps=True,
-	glyphs_optimize_names=glyphs_optimize_names,
-	glyphs_omit_list=glyphs_omit_list,
-	glyphs_omit_suffixes_list=glyphs_omit_suffixes_list,
-	groups_flc_path=flc_path,
-	)
+  glyphs_decompose=True,
+  glyphs_remove_overlaps=True,
+  glyphs_optimize_names=glyphs_optimize_names,
+  glyphs_omit_list=glyphs_omit_list,
+  glyphs_omit_suffixes_list=glyphs_omit_suffixes_list,
+  groups_flc_path=flc_path,
+  )
 ```
 
 Test (~2900 glyphs @ 10,000 UPM -> 1,000 UPM), ≈4 sec
@@ -605,9 +616,9 @@ Test (~2900 glyphs @ 10,000 UPM -> 1,000 UPM), ≈4 sec
 ```
 flc_path = <path to .flc file>
 vfb2ufo3.write_ufo(
-	glyphs_decompose=True,
-	groups_flc_path=flc_path,
-	)
+  glyphs_decompose=True,
+  groups_flc_path=flc_path,
+  )
 ```
 
 Test (~2900 glyphs @ 10,000 UPM -> 1,000 UPM), ≈3-4 sec
@@ -615,18 +626,18 @@ Test (~2900 glyphs @ 10,000 UPM -> 1,000 UPM), ≈3-4 sec
 ```
 flc_path = <path to .flc file>
 glyphs_omit_list = [
-	<glyph names to be omitted go here>
-	]
+  <glyph names to be omitted go here>
+  ]
 glyphs_omit_suffixes_list = [
-	<glyph name suffixes to be omitted go here>
-	]
+  <glyph name suffixes to be omitted go here>
+  ]
 
 vfb2ufo3.write_ufo(
-	glyphs_decompose=True,
-	glyphs_omit_list=glyphs_omit_list,
-	glyphs_omit_suffixes_list=glyphs_omit_suffixes_list,
-	groups_flc_path=flc_path,
-	)
+  glyphs_decompose=True,
+  glyphs_omit_list=glyphs_omit_list,
+  glyphs_omit_suffixes_list=glyphs_omit_suffixes_list,
+  groups_flc_path=flc_path,
+  )
 ```
 
 Test (~3200 glyphs @ 10,000 UPM -> 1,000 UPM), ≈3 sec
@@ -635,8 +646,8 @@ Test (~3200 glyphs @ 10,000 UPM -> 1,000 UPM), ≈3 sec
 flc_path = <path to .flc file>
 
 vfb2ufo3.write_ufo(
-	groups_flc_path=flc_path,
-	)
+  groups_flc_path=flc_path,
+  )
 ```
 
 Test (~3200 glyphs @ 10,000 UPM -> 1,000 UPM), <1.5 sec
@@ -645,9 +656,9 @@ Test (~3200 glyphs @ 10,000 UPM -> 1,000 UPM), <1.5 sec
 flc_path = <path to .flc file>
 
 vfb2ufo3.write_ufo(
-	groups_flc_path=flc_path,
-	ufoz=True,
-	)
+  groups_flc_path=flc_path,
+  ufoz=True,
+  )
 ```
 
 Test (~3200 glyphs @ 10,000 UPM -> 1,000 UPM), <1.5 sec
@@ -656,10 +667,10 @@ Test (~3200 glyphs @ 10,000 UPM -> 1,000 UPM), <1.5 sec
 flc_path = <path to .flc file>
 
 vfb2ufo3.write_ufo(
-	groups_flc_path=flc_path,
-	ufoz=True,
-	ufoz_compress=False,
-	)
+  groups_flc_path=flc_path,
+  ufoz=True,
+  ufoz_compress=False,
+  )
 ```
 
 Notes
@@ -673,9 +684,22 @@ Author
 Jameson R Spires
 
 Licenses
-Source files are covered under the MIT License.
+Source files are covered under the MIT License. SHA512 hash algorithm is a
+modification of an implementation from Olivier Gay's SHA2 library, which is
+covered under a BSD License.
 
 Version history
+version 0.8.1
+added `glyphs_optimize_makeotf` option for use with makeOTF
+added `glyphs_decompose_names` and `glyphs_remove_overlap_names` options
+correction to SHA512 hint ID hashing
+correction to font groups parsing when not using `.flc` or `groups.plist` group
+files
+correction to features.fea copy
+UFO glyph order derivation from encoding now checks for and includes glyphs
+missing from FontLab `.enc` file
+removed option to save VFB instances/copies
+
 version 0.8.0
 a copy of the original font is no longer created, decreasing master processing
 overhead by a considerable amount
@@ -771,101 +795,103 @@ initial release
 """
 
 def cleanup():
-	gc.collect()
-	del gc.garbage[:]
+  gc.collect()
+  del gc.garbage[:]
 
 def write_ufo(
-	output_path=None,
+  output_path=None,
 
-	layer=None,
+  layer=None,
 
-	scale_to_upm=1000,
-	scale_auto=True,
+  scale_to_upm=1000,
+  scale_auto=True,
 
-	instance_values=[],
-	instance_names=[],
-	instance_attributes=[],
+  instance_values=[],
+  instance_names=[],
+  instance_attributes=[],
 
-	features_import_groups=False,
+  features_import_groups=False,
 
-	kern_feature_generate=True,
-	kern_feature_file_path=None,
-	kern_feature_passthrough=False,
-	kern_min_value=None,
+  kern_feature_generate=True,
+  kern_feature_file_path=None,
+  kern_feature_passthrough=False,
+  kern_min_value=None,
 
-	mark_feature_generate=False,
-	mark_anchors_include=[],
-	mark_anchors_omit=[],
+  mark_feature_generate=False,
+  mark_anchors_include=[],
+  mark_anchors_omit=[],
 
-	groups_export_flc=False,
-	groups_export_flc_path=None,
-	groups_flc_path=None,
-	groups_plist_path=None,
-	groups_ignore_no_kerning=False,
+  groups_export_flc=False,
+  groups_export_flc_path=None,
+  groups_flc_path=None,
+  groups_plist_path=None,
+  groups_ignore_no_kerning=False,
 
-	afdko_parts=False,
-	afdko_makeotf_cmd=False,
-	afdko_makeotf_batch_cmd=False,
-	afdko_makeotf_output_dir=None,
-	afdko_makeotf_GOADB_path=None,
-	afdko_makeotf_GOADB_win1252=True,
-	afdko_makeotf_GOADB_macos_roman=False,
-	afdko_makeotf_release=False,
-	afdko_makeotf_subroutinization=True,
-	afdko_makeotf_no_subroutinization=False,
-	afdko_makeotf_sans=False,
-	afdko_makeotf_serif=False,
-	afdko_makeotf_replace_notdef=False,
-	afdko_makeotf_verbose=False,
-	afdko_makeotf_addDSIG=True,
-	afdko_makeotf_suppress_unhinted_glyph_warnings=True,
-	afdko_makeotf_args=[],
+  afdko_parts=False,
+  afdko_makeotf_cmd=False,
+  afdko_makeotf_batch_cmd=False,
+  afdko_makeotf_output_dir=None,
+  afdko_makeotf_GOADB_path=None,
+  afdko_makeotf_GOADB_win1252=True,
+  afdko_makeotf_GOADB_macos_roman=False,
+  afdko_makeotf_release=False,
+  afdko_makeotf_subroutinization=True,
+  afdko_makeotf_no_subroutinization=False,
+  afdko_makeotf_sans=False,
+  afdko_makeotf_serif=False,
+  afdko_makeotf_replace_notdef=False,
+  afdko_makeotf_verbose=False,
+  afdko_makeotf_addDSIG=True,
+  afdko_makeotf_suppress_unhinted_glyph_warnings=True,
+  afdko_makeotf_args=[],
 
-	psautohint_cmd=False,
-	psautohint_batch_cmd=False,
-	psautohint_write_to_default_layer=True,
-	psautohint_decimal=True,
-	psautohint_allow_outline_changes=False,
-	psautohint_no_flex=False,
-	psautohint_no_hint_substitution=False,
-	psautohint_no_zones_stems=False,
-	psautohint_log=False,
-	psautohint_report_only=False,
-	psautohint_verbose=False,
-	psautohint_extra_verbose=False,
-	psautohint_glyphs_list=[],
-	psautohint_glyphs_omit_list=[],
+  psautohint_cmd=False,
+  psautohint_batch_cmd=False,
+  psautohint_write_to_default_layer=True,
+  psautohint_decimal=True,
+  psautohint_allow_outline_changes=False,
+  psautohint_no_flex=False,
+  psautohint_no_hint_substitution=False,
+  psautohint_no_zones_stems=False,
+  psautohint_log=False,
+  psautohint_report_only=False,
+  psautohint_verbose=False,
+  psautohint_extra_verbose=False,
+  psautohint_glyphs_list=[],
+  psautohint_glyphs_omit_list=[],
 
-	glyphs_decompose=False,
-	glyphs_remove_overlaps=False,
+  glyphs_decompose=False,
+  glyphs_remove_overlaps=False,
 
-	glyphs_omit_names=[],
-	glyphs_omit_suffixes=[],
-	glyphs_optimize=True,
-	glyphs_optimize_code_points=[],
-	glyphs_optimize_names=[],
+  glyphs_omit_names=[],
+  glyphs_omit_suffixes=[],
+  glyphs_optimize=True,
+  glyphs_optimize_makeotf=False,
+  glyphs_optimize_code_points=[],
+  glyphs_optimize_names=[],
 
-	glyphs_hints=False,
-	glyphs_hints_vertical_only=False,
-	glyphs_hints_afdko_v1=False,
-	glyphs_hints_afdko_v2=False,
+  glyphs_hints=False,
+  glyphs_hints_vertical_only=False,
+  glyphs_hints_afdko_v1=False,
+  glyphs_hints_afdko_v2=False,
 
-	ufoz=False,
-	ufoz_compress=True,
+  ufoz=False,
+  ufoz_compress=True,
 
-	designspace_export=False,
-	designspace_default=[],
+  designspace_export=False,
+  designspace_default=[],
 
-	vfb_save=False,
-	vfb_close=True,
+  force_overwrite=False,
 
-	force_overwrite=False,
+  report=True,
+  report_verbose=False,
+  ):
 
-	report=True,
-	report_verbose=False,
-	):
+  if not len(fl):
+    raise UserWarning(b'No open fonts')
 
-	options = core.decode_dict(locals())
-	core.write_ufo(options)
-	cleanup()
+  fl.output = b''
+  options = core.decode_dict(locals())
+  core.write_ufo(options)
+  cleanup()
 
